#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dialogs.database.py
"""
Classes based on wx.Dialog for models.database_editor.py
In general, this classes will manifest themselves in the GUI on initialization.  They should contain a function names
action which will be executed on a dialog resolution of wx.ID_OK
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from datetime import datetime
from os import mkdir, rename
from os.path import join, basename
from dvha.db.sql_connector import DVH_SQL, echo_sql_db, is_file_sqlite_db
from dvha.models.import_dicom import ImportDicomFrame
from dvha.paths import DATA_DIR
from dvha.tools.errors import SQLError, SQLErrorDialog
from dvha.tools.utilities import delete_directory_contents, move_files_to_new_path, delete_file, get_file_paths,\
    delete_imported_dicom_files, move_imported_dicom_files, MessageDialog


class CalculationsDialog(wx.Dialog):
    # TODO: Add functionality to OK button
    """
    Dialog to perform various calculations for values not stored in DICOM files.
    """
    def __init__(self):
        wx.Dialog.__init__(self, None, title="Calculations")

        choices = ["PTV Distances", "PTV Overlap", "ROI Centroid", "ROI Spread", "ROI Cross-Section",
                   "OAR-PTV Centroid Distance", "Beam Complexities", "Plan Complexities", "All (except age)",
                   "Patient Ages"]
        self.combo_box_calculate = wx.ComboBox(self, wx.ID_ANY, choices=choices, style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.checkbox = wx.CheckBox(self, wx.ID_ANY, "Only Calculate Missing Values")
        self.text_ctrl_condition = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "Calculate")
        self.button_ok.Disable()  # Remove this line once functionality is added
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.checkbox.SetValue(1)

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_wrapper_inner = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL)
        sizer_condition = wx.BoxSizer(wx.VERTICAL)
        sizer_calc_and_check = wx.BoxSizer(wx.HORIZONTAL)
        sizer_calculate = wx.BoxSizer(wx.VERTICAL)

        label_calculate = wx.StaticText(self, wx.ID_ANY, "Calculate:")
        sizer_calculate.Add(label_calculate, 0, wx.BOTTOM, 5)
        sizer_calculate.Add(self.combo_box_calculate, 0, 0, 0)
        sizer_calc_and_check.Add(sizer_calculate, 0, wx.EXPAND, 0)
        sizer_calc_and_check.Add(self.checkbox, 0, wx.LEFT | wx.TOP, 20)
        sizer_input.Add(sizer_calc_and_check, 0, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5)

        label_condition = wx.StaticText(self, wx.ID_ANY, "Condition:")
        sizer_condition.Add(label_condition, 0, wx.BOTTOM, 5)
        sizer_condition.Add(self.text_ctrl_condition, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_condition, 0, wx.ALL | wx.EXPAND, 5)
        sizer_wrapper_inner.Add(sizer_input, 0, wx.ALL | wx.EXPAND, 5)

        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper_inner.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_wrapper_inner, 0, wx.EXPAND, 0)

        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            pass
        self.Destroy()


class ChangeOrDeleteBaseClass(wx.Dialog):
    """
    A class generalized for changing a patient identifier or deleting a study for a patient
    """
    def __init__(self, text_input_1_label, text_input_2_label, ok_button_label, title,
                 mrn=None, study_instance_uid=None):
        """
        :param text_input_1_label: label of first text input in GUI
        :type text_input_1_label: str
        :param text_input_2_label: label of second text input in GUI
        :type text_input_2_label: str
        :param ok_button_label: label of OK button in GUI
        :type ok_button_label: str
        :param title: title of the wx.Dialog object
        :type title: str
        :param mrn: optional initial value for mrn
        :type mrn: str
        :param study_instance_uid: options initial value for study instance uid
        :type study_instance_uid: str
        """
        wx.Dialog.__init__(self, None, title=title)

        self.initial_mrn = mrn
        self.initial_study_instance_uid = study_instance_uid

        self.text_input_1_label = text_input_1_label
        self.text_input_2_label = text_input_2_label

        self.combo_box_patient_identifier = wx.ComboBox(self, wx.ID_ANY, choices=["MRN", "Study Instance UID"],
                                                        style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.text_ctrl_1 = wx.TextCtrl(self, wx.ID_ANY, "")
        self.text_ctrl_2 = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, ok_button_label)
        self.button_ok.Disable()
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.text_ctrl_1.SetMinSize((365, 22))
        if self.initial_mrn or self.initial_study_instance_uid is None:
            self.combo_box_patient_identifier.SetValue('MRN')
        else:
            self.combo_box_patient_identifier.SetValue('Study Instance UID')
        self.on_identifier_change(None)

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL)
        sizer_new_value = wx.BoxSizer(wx.VERTICAL)
        sizer_value = wx.BoxSizer(wx.VERTICAL)
        sizer_patient_identifier = wx.BoxSizer(wx.HORIZONTAL)
        label_patient_identifier = wx.StaticText(self, wx.ID_ANY, "Patient Identifier:")
        sizer_patient_identifier.Add(label_patient_identifier, 0, wx.ALL, 5)
        sizer_patient_identifier.Add(self.combo_box_patient_identifier, 0, wx.TOP, 2)
        sizer_input.Add(sizer_patient_identifier, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        label_text_input_1 = wx.StaticText(self, wx.ID_ANY, self.text_input_1_label)
        sizer_value.Add(label_text_input_1, 0, wx.EXPAND | wx.ALL, 5)
        sizer_value.Add(self.text_ctrl_1, 0, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_input.Add(sizer_value, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        label_text_input_2 = wx.StaticText(self, wx.ID_ANY, self.text_input_2_label)
        sizer_new_value.Add(label_text_input_2, 0, wx.EXPAND | wx.ALL, 5)
        sizer_new_value.Add(self.text_ctrl_2, 0, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_input.Add(sizer_new_value, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_wrapper.Add(sizer_input, 0, wx.LEFT | wx.RIGHT | wx.TOP, 10)
        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.BOTTOM | wx.RIGHT, 10)
        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    def __do_bind(self):
        self.Bind(wx.EVT_COMBOBOX, self.on_identifier_change, id=self.combo_box_patient_identifier.GetId())
        self.Bind(wx.EVT_TEXT, self.text_ticker, id=self.text_ctrl_1.GetId())
        self.Bind(wx.EVT_TEXT, self.text_ticker, id=self.text_ctrl_2.GetId())

    def text_ticker(self, evt):
        [self.button_ok.Disable, self.button_ok.Enable][self.is_action_allowed]()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        """
        Should be over-written in classes that inherit this class, but needs to be defined here since run refers to it
        """
        pass

    def on_identifier_change(self, evt):
        value = {'Study Instance UID': self.initial_study_instance_uid,
                 'MRN': self.initial_mrn}[self.combo_box_patient_identifier.GetValue()]
        if value is not None:
            self.text_ctrl_1.SetValue(value)
            wx.CallAfter(self.text_ctrl_2.SetFocus)

    @property
    def sql_column(self):
        return self.combo_box_patient_identifier.GetValue().lower().replace(' ', '_')

    @property
    def is_id_valid(self):
        with DVH_SQL() as cnx:
            func = [cnx.is_uid_imported, cnx.is_mrn_imported][self.combo_box_patient_identifier.GetValue() == 'MRN']
            ans = func(self.text_ctrl_1.GetValue())
        return ans

    @property
    def is_action_allowed(self):
        return self.is_id_valid and bool(self.text_ctrl_2.GetValue())


class ChangePatientIdentifierDialog(ChangeOrDeleteBaseClass):
    """
    Change MRN or Study Instance UID in all SQL tables
    """
    def __init__(self, mrn=None, study_instance_uid=None):
        """
        :param mrn: optional initial value for mrn
        :type mrn: str
        :param study_instance_uid: options initial value for study instance uid
        :type study_instance_uid: str
        """
        ChangeOrDeleteBaseClass.__init__(self, 'Value:', 'New Value:', 'Change', "Change Patient Identifier",
                                         mrn=mrn, study_instance_uid=study_instance_uid)

    def action(self):
        old_id = self.text_ctrl_1.GetValue()
        new_id = self.text_ctrl_2.GetValue()

        with DVH_SQL() as cnx:
            validation_func = [cnx.is_uid_imported, cnx.is_mrn_imported][self.sql_column == 'mrn']
            change_func = [cnx.change_uid, cnx.change_mrn][self.sql_column == 'mrn']

            if validation_func(old_id):
                if self.sql_column == 'study_instance_uid' and validation_func(new_id):
                    wx.MessageBox('This Study Instance UID is already in use.',
                                  '%s Error' % self.combo_box_patient_identifier.GetValue(),
                                  wx.OK | wx.ICON_WARNING)
                else:
                    change_func(old_id, new_id)
            else:
                wx.MessageBox('No studies found with this %s.' % self.combo_box_patient_identifier.GetValue(),
                              '%s Error' % self.combo_box_patient_identifier.GetValue(),
                              wx.OK | wx.ICON_WARNING)


class DeletePatientDialog(ChangeOrDeleteBaseClass):
    """
    Delete all data in SQL database for a given MRN or study instance uid
    """
    def __init__(self, mrn=None, study_instance_uid=None):
        """
        :param mrn: optional initial value for mrn
        :type mrn: str
        :param study_instance_uid: options initial value for study instance uid
        :type study_instance_uid: str
        """
        ChangeOrDeleteBaseClass.__init__(self, 'Delete:', 'Type "delete" to authorize:', 'Delete', "Delete Patient",
                                         mrn=mrn, study_instance_uid=study_instance_uid)

    def action(self):
        if self.text_ctrl_2.GetValue() == 'delete':
            value = self.text_ctrl_1.GetValue()
            with DVH_SQL() as cnx:
                dicom_files = cnx.get_dicom_file_paths(**{self.sql_column: value})
                cnx.delete_rows("%s = '%s'" % (self.sql_column, value))

            DeleteFilesFromQuery(self, dicom_files)

    @property
    def is_action_allowed(self):
        return self.is_id_valid and self.text_ctrl_2.GetValue() == 'delete'


class EditDatabaseDialog(wx.Dialog):
    """
    Generic UI object to edit values in SQL database with table, column, and condition
    """
    def __init__(self, inital_values=None):
        """
        :param inital_values: optional variable to provide initial values to table, column, value, or condition
        :type inital_values: dict
        """
        wx.Dialog.__init__(self, None, title="Edit Database Values")

        self.initial_values = inital_values
        self.error = False

        self.combo_box_table = wx.ComboBox(self, wx.ID_ANY, choices=self.tables,
                                           style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.combo_box_column = wx.ComboBox(self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.text_ctrl_value = wx.TextCtrl(self, wx.ID_ANY, "")
        self.text_ctrl_condition = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "Update")
        self.button_ok.Disable()
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.Bind(wx.EVT_COMBOBOX, self.table_ticker, id=self.combo_box_table.GetId())
        self.Bind(wx.EVT_TEXT, self.condition_ticker, id=self.text_ctrl_condition.GetId())

        self.__set_properties()
        self.__do_layout()

        self.update_columns()

        self.run()

    def __set_properties(self):
        if self.initial_values:
            self.combo_box_table.SetValue(self.initial_values['table'])
            self.combo_box_column.SetValue(self.initial_values['column'])
            self.text_ctrl_value.SetValue(self.initial_values['value'])
            self.text_ctrl_condition.SetValue(self.initial_values['condition'])
        else:
            self.combo_box_table.SetValue('Plans')

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_wrapper_inner = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL)
        sizer_condition = wx.BoxSizer(wx.VERTICAL)
        sizer_table_column_value = wx.BoxSizer(wx.HORIZONTAL)
        sizer_value = wx.BoxSizer(wx.VERTICAL)
        sizer_column = wx.BoxSizer(wx.VERTICAL)
        sizer_table = wx.BoxSizer(wx.VERTICAL)
        label_table = wx.StaticText(self, wx.ID_ANY, "Table:")
        sizer_table.Add(label_table, 0, 0, 0)
        sizer_table.Add(self.combo_box_table, 0, wx.EXPAND, 0)
        sizer_table_column_value.Add(sizer_table, 0, wx.ALL | wx.EXPAND, 5)
        label_column = wx.StaticText(self, wx.ID_ANY, "Column:")
        sizer_column.Add(label_column, 0, 0, 0)
        sizer_column.Add(self.combo_box_column, 0, 0, 0)
        sizer_table_column_value.Add(sizer_column, 0, wx.ALL | wx.EXPAND, 5)
        label_value = wx.StaticText(self, wx.ID_ANY, "Value:")
        sizer_value.Add(label_value, 0, 0, 0)
        sizer_value.Add(self.text_ctrl_value, 0, wx.EXPAND, 0)
        sizer_table_column_value.Add(sizer_value, 1, wx.ALL | wx.EXPAND, 5)
        sizer_input.Add(sizer_table_column_value, 1, wx.EXPAND, 0)
        label_condition = wx.StaticText(self, wx.ID_ANY, "Condition:")
        sizer_condition.Add(label_condition, 0, wx.BOTTOM | wx.RIGHT | wx.TOP, 5)
        sizer_condition.Add(self.text_ctrl_condition, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_condition, 1, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_wrapper_inner.Add(sizer_input, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper_inner.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        sizer_wrapper.Add(sizer_wrapper_inner, 0, wx.ALL | wx.EXPAND, 0)
        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    @property
    def tables(self):
        with DVH_SQL() as cnx:
            tables = cnx.tables
        return tables

    @property
    def table(self):
        return self.combo_box_table.GetValue()

    @property
    def column(self):
        return self.combo_box_column.GetValue()

    @property
    def condition(self):
        return self.text_ctrl_condition.GetValue()

    @property
    def value(self):
        v = self.text_ctrl_value.GetValue()
        if v:
            return v
        return 'NULL'

    def condition_ticker(self, evt):
        if self.condition:
            self.button_ok.Enable()
        else:
            self.button_ok.Disable()

    def update_columns(self):
        table = self.combo_box_table.GetValue()
        with DVH_SQL() as cnx:
            columns = cnx.get_column_names(table)
        self.combo_box_column.SetItems(columns)
        if self.combo_box_column.GetValue() not in columns:
            self.combo_box_column.SetValue(columns[0])

    def table_ticker(self, evt):
        self.update_columns()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            self.update_db()
        self.Destroy()
        if self.error:
            EditDatabaseDialog(inital_values=self.values)

    def update_db(self):
        with DVH_SQL() as cnx:
            try:
                cnx.update(self.table, self.column, self.value, self.condition)
            except SQLError as sql_exception:
                SQLErrorDialog(self, sql_exception)
                self.error = True

    @property
    def values(self):
        return {'table': self.combo_box_table.GetValue(),
                'column': self.combo_box_column.GetValue(),
                'value': self.text_ctrl_value.GetValue(),
                'condition': self.text_ctrl_condition.GetValue()}


class ReimportDialog(wx.Dialog):
    """
    Reimport data from catalogued DICOM files for a given MRN or study instance uid
    """
    def __init__(self, roi_map, options, mrn=None, study_instance_uid=None):
        """
        :param mrn: optional initial value for mrn
        :type mrn: str
        :param study_instance_uid: options initial value for study instance uid
        :type study_instance_uid: str
        """
        wx.Dialog.__init__(self, None, title="Reimport from DICOM")
        self.roi_map = roi_map
        self.options = options
        self.initial_mrn = mrn
        self.initial_uid = study_instance_uid

        self.text_ctrl_mrn = wx.TextCtrl(self, wx.ID_ANY, "")
        self.radio_box_delete_from_db = wx.RadioBox(self, wx.ID_ANY, "Current Data",
                                                    choices=["Delete from DB", "Keep in DB"],
                                                    majorDimension=2, style=wx.RA_SPECIFY_ROWS)
        self.combo_box_study_date = wx.ComboBox(self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.combo_box_uid = wx.ComboBox(self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.button_reimport = wx.Button(self, wx.ID_OK, "Reimport")
        self.button_reimport.Disable()
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_layout()
        self.__do_bind()
        self.__apply_initial_values()

        self.run()

    def __set_properties(self):
        self.radio_box_delete_from_db.SetSelection(0)
        self.combo_box_study_date.SetMinSize((200, 25))

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.mrn_ticker, id=self.text_ctrl_mrn.GetId())
        self.Bind(wx.EVT_COMBOBOX, self.study_date_ticker, id=self.combo_box_study_date.GetId())
        self.Bind(wx.EVT_COMBOBOX, self.uid_ticker, id=self.combo_box_uid.GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL)
        sizer_input_date_uid = wx.BoxSizer(wx.HORIZONTAL)
        sizer_uid = wx.BoxSizer(wx.VERTICAL)
        sizer_input_date = wx.BoxSizer(wx.VERTICAL)
        sizer_input_mrn_db = wx.BoxSizer(wx.HORIZONTAL)
        sizer_mrn = wx.BoxSizer(wx.VERTICAL)
        label_mrn = wx.StaticText(self, wx.ID_ANY, "MRN:")
        sizer_mrn.Add(label_mrn, 0, wx.BOTTOM, 5)
        sizer_mrn.Add(self.text_ctrl_mrn, 0, wx.EXPAND | wx.RIGHT, 40)
        sizer_input_mrn_db.Add(sizer_mrn, 1, wx.TOP, 12)
        sizer_input_mrn_db.Add(self.radio_box_delete_from_db, 0, wx.ALL, 5)
        sizer_input.Add(sizer_input_mrn_db, 0, wx.EXPAND | wx.LEFT, 5)
        label_date = wx.StaticText(self, wx.ID_ANY, "Sim Study Date:")
        sizer_input_date.Add(label_date, 0, wx.BOTTOM, 5)
        sizer_input_date.Add(self.combo_box_study_date, 0, 0, 0)
        sizer_input_date_uid.Add(sizer_input_date, 0, wx.ALL | wx.EXPAND, 5)
        label_uid = wx.StaticText(self, wx.ID_ANY, "Study Instance UID:")
        sizer_uid.Add(label_uid, 0, wx.BOTTOM, 5)
        sizer_uid.Add(self.combo_box_uid, 0, wx.EXPAND, 0)
        sizer_input_date_uid.Add(sizer_uid, 1, wx.ALL | wx.EXPAND, 5)
        sizer_input.Add(sizer_input_date_uid, 0, wx.EXPAND, 0)
        sizer_wrapper.Add(sizer_input, 0, wx.ALL | wx.EXPAND, 5)
        sizer_ok_cancel.Add(self.button_reimport, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.BOTTOM | wx.RIGHT, 5)
        self.SetSizer(sizer_wrapper)
        sizer_wrapper.SetMinSize((700, 190))
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    def __apply_initial_values(self):
        if self.initial_mrn is not None:
            self.text_ctrl_mrn.SetValue(self.initial_mrn)
        # if self.initial_uid is not None:
        #     self.combo_box_uid.SetSelection(self.initial_uid)

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            if self.uid:
                self.action()
        self.Destroy()

    def mrn_ticker(self, evt):
        with DVH_SQL() as cnx:
            is_mrn_valid = cnx.is_mrn_imported(self.mrn)
        if is_mrn_valid:
            self.update_study_dates()
        else:
            self.combo_box_study_date.SetItems([])
            self.combo_box_uid.SetItems([])
            self.button_reimport.Disable()

    def study_date_ticker(self, evt):
        self.update_uids()

    def uid_ticker(self, evt):
        if self.combo_box_uid.GetValue():
            self.button_reimport.Enable()
        else:
            self.button_reimport.Disable()

    def update_study_dates(self):
        with DVH_SQL() as cnx:
            choices = cnx.get_unique_values('Plans', 'sim_study_date', "mrn = '%s'" % self.mrn)
        self.combo_box_study_date.SetItems(choices)
        if choices:
            self.combo_box_study_date.SetValue(choices[0])
        self.update_uids()

    def update_uids(self):

        date = ['is NULL', "= '%s'::date" % self.sim_study_date][self.sim_study_date != 'None']
        with DVH_SQL() as cnx:
            if cnx.db_type == 'sqlite':
                date = date.replace('::date', '')
            condition = "mrn = '%s' and sim_study_date %s" % (self.mrn, date)
            choices = cnx.get_unique_values('Plans', 'study_instance_uid', condition)
        self.combo_box_uid.SetItems(choices)
        if choices:
            self.combo_box_uid.SetValue(choices[0])
        else:
            self.combo_box_uid.SetValue(None)
        self.uid_ticker(None)

    @property
    def delete_from_db(self):
        return self.radio_box_delete_from_db.GetSelection() == 0

    @property
    def mrn(self):
        return self.text_ctrl_mrn.GetValue()

    @property
    def sim_study_date(self):
        return self.combo_box_study_date.GetValue()

    @property
    def uid(self):
        return self.combo_box_uid.GetValue()

    def action(self):
        with DVH_SQL() as cnx:
            dicom_files = cnx.get_dicom_file_paths(uid=self.uid)
            if self.delete_from_db:
                cnx.delete_rows("study_instance_uid = '%s'" % self.uid)
        inbox = self.options.INBOX_DIR
        move_imported_dicom_files(dicom_files, inbox)
        ImportDicomFrame(self.roi_map, self.options, inbox=inbox, auto_parse=True)


class SQLSettingsDialog(wx.Dialog):
    """
    Edit and validate SQL connection settings
    """
    def __init__(self, options):
        wx.Dialog.__init__(self, None, title="Database Connection Settings")

        self.options = options

        self.keys = ['host', 'port', 'dbname', 'user', 'password']

        self.input = {'host': wx.ComboBox(self, wx.ID_ANY, "")}
        for key in self.keys:
            if key not in {'password', 'host'}:
                self.input[key] = wx.TextCtrl(self, wx.ID_ANY, "")
        self.input['password'] = wx.TextCtrl(self, wx.ID_ANY, "", style=wx.TE_PASSWORD)

        self.db_type_radiobox = wx.RadioBox(self, wx.ID_ANY, 'Database Type', choices=['SQLite', 'Postgres'])

        self.button = {'echo': wx.Button(self, wx.ID_ANY, "Echo"),
                       'reload': wx.Button(self, wx.ID_ANY, "Reload Last Connection"),
                       'ok': wx.Button(self, wx.ID_OK, "OK"),
                       'cancel': wx.Button(self, wx.ID_CANCEL, "Cancel")}

        self.db_types = ['sqlite', 'pgsql']

        self.ip_history = self.options.SQL_PGSQL_IP_HIST

        self.Bind(wx.EVT_BUTTON, self.button_echo, id=self.button['echo'].GetId())
        self.Bind(wx.EVT_BUTTON, self.button_reload, id=self.button['reload'].GetId())
        self.Bind(wx.EVT_RADIOBOX, self.on_db_radio, id=self.db_type_radiobox.GetId())

        self.__set_properties()
        self.__do_layout()
        self.Center()

        self.load_sql_settings()

        self.run()

    def __set_properties(self):
        self.SetTitle("SQL Connection Settings")

        # Set initial db_type_radiobox to loaded settings or pgsql if none found
        self.set_selected_db_type(self.options.DB_TYPE)

        self.button['reload'].Enable(self.has_last_cnx)

    def set_host_items(self):
        if self.selected_db_type == 'sqlite':
            db_files = [basename(f) for f in get_file_paths(DATA_DIR, extension='.db') if is_file_sqlite_db(f)]
            db_files.sort()
            self.input['host'].SetItems(db_files)
        else:
            self.input['host'].SetItems(self.ip_history)

    def __do_layout(self):
        sizer_frame = wx.BoxSizer(wx.VERTICAL)
        sizer_echo = wx.BoxSizer(wx.HORIZONTAL)
        sizer_reload = wx.BoxSizer(wx.HORIZONTAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        grid_sizer = wx.GridSizer(5, 2, 5, 10)

        label_titles = ['Host:', 'Port:', 'Database Name:', 'User Name:', 'Password:']
        self.label = {key: wx.StaticText(self, wx.ID_ANY, label_titles[i]) for i, key in enumerate(self.keys)}

        for key in self.keys:
            grid_sizer.Add(self.label[key], 0, wx.ALL, 0)
            grid_sizer.Add(self.input[key], 0, wx.ALL | wx.EXPAND, 0)

        sizer_frame.Add(grid_sizer, 0, wx.ALL, 10)
        sizer_frame.Add(self.db_type_radiobox, 0, wx.EXPAND | wx.ALL, 5)
        sizer_buttons.Add(self.button['ok'], 1, wx.ALL | wx.EXPAND, 5)
        sizer_buttons.Add(self.button['cancel'], 1, wx.ALL | wx.EXPAND, 5)
        sizer_echo.Add(self.button['echo'], 1, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        sizer_frame.Add(sizer_echo, 1, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        sizer_reload.Add(self.button['reload'], 1, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        sizer_frame.Add(sizer_reload, 1, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        sizer_frame.Add(sizer_buttons, 0, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_frame)
        self.Fit()
        self.Layout()

    def load_sql_settings(self):
        self.set_host_items()

        if self.has_last_cnx:
            config = self.options.SQL_LAST_CNX[self.selected_db_type]
        else:
            config = self.options.DEFAULT_CNF[self.selected_db_type]

        self.clear_input()

        if self.selected_db_type == 'sqlite':
            self.input['host'].SetValue(config['host'])
        else:
            for input_type in self.keys:
                if input_type in config:
                    self.input[input_type].SetValue(config[input_type])

    def button_echo(self, evt):
        if self.valid_sql_settings:
            wx.MessageBox('Success!', 'Echo SQL Database', wx.OK | wx.ICON_INFORMATION)
        else:
            wx.MessageBox('Invalid credentials!', 'Echo SQL Database', wx.OK | wx.ICON_WARNING)

    def button_reload(self, evt):
        self.load_sql_settings()

    def write_successful_cnf(self):
        new_config = {key: self.input[key].GetValue() for key in self.keys if self.input[key].GetValue()}
        self.options.SQL_LAST_CNX[self.selected_db_type] = new_config

        self.options.DB_TYPE = self.selected_db_type

        if self.selected_db_type == 'pgsql':
            new_host = self.input['host'].GetValue()
            if new_host in self.ip_history:
                self.ip_history.pop(self.ip_history.index(new_host))
            self.ip_history.insert(0, new_host)

        self.options.save()

    @property
    def valid_sql_settings(self):
        config = {key: self.input[key].GetValue() for key in self.keys if self.input[key].GetValue()}
        if self.selected_db_type == 'pgsql':
            config['dbname'] = self.input['dbname'].GetValue()
        return echo_sql_db(config, db_type=self.selected_db_type)

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            new_config = {key: self.input[key].GetValue() for key in self.keys if self.input[key].GetValue()}

            if echo_sql_db(new_config, db_type=self.selected_db_type):
                self.write_successful_cnf()
                with DVH_SQL() as cnx:
                    cnx.initialize_database()
            else:
                dlg = wx.MessageDialog(self, 'Connection to database could not be established.', 'ERROR!',
                                       wx.OK | wx.ICON_ERROR)
                dlg.ShowModal()
        self.Destroy()

    def on_db_radio(self, *evt):
        self.load_sql_settings()
        self.button['reload'].Enable(self.has_last_cnx)

    @property
    def selected_db_type(self):
        return ['sqlite', 'pgsql'][self.db_type_radiobox.GetSelection()]

    @property
    def unselected_db_type(self):
        return ['pgsql', 'sqlite'][self.db_type_radiobox.GetSelection()]

    @property
    def has_last_cnx(self):
        return 'host' in list(self.options.SQL_LAST_CNX[self.selected_db_type])

    def set_selected_db_type(self, db_type):
        self.db_type_radiobox.SetSelection({'sqlite': 0, 'pgsql': 1}[db_type])

    def clear_input(self):
        for input_type in self.keys:
            self.input[input_type].SetValue('')
            if input_type != 'host':
                self.input[input_type].Enable(self.selected_db_type == 'pgsql')
                self.label[input_type].Enable(self.selected_db_type == 'pgsql')


# ------------------------------------------------------
# Yes/No Dialogs based on tools.utilities.MessageDialog
#
# MessageDialog handles the layout and display, simply
# provide the wx parent, a message, and define over-ride
# action_yes and/or action_no for Yes/No button clicks
# ------------------------------------------------------
class DeleteAllData(MessageDialog):
    def __init__(self, parent, options):
        self.options = options
        MessageDialog.__init__(self, parent, "Delete All Data in Database")

    def action_yes(self):
        with DVH_SQL() as cnx:
            cnx.reinitialize_database()
        MoveFilesToInbox(self.parent, self.options.INBOX_DIR, self.options.IMPORTED_DIR)


class MoveFilesToInbox(MessageDialog):
    def __init__(self, parent, inbox_dir, imported_dir):
        self.parent = parent
        self.inbox_dir = inbox_dir
        self.imported_dir = imported_dir
        MessageDialog.__init__(self, parent, "Move files to inbox?")

    def action_yes(self):
        new_dir = join(self.inbox_dir, "previously_imported %s" %
                       str(datetime.now()).split('.')[0].replace(':', '-').replace(' ', '_'))
        rename(self.imported_dir, new_dir)
        mkdir(self.imported_dir)

    def action_no(self):
        DeleteImportedDirectory(self.parent, self.imported_dir)


class DeleteImportedDirectory(MessageDialog):
    def __init__(self, parent, directory):
        self.directory = directory
        MessageDialog.__init__(self, parent, "Delete Imported Directory?")

    def action_yes(self):
        delete_directory_contents(self.directory)


class MoveFiles(MessageDialog):
    def __init__(self, parent, files):
        self.parent = parent
        self.files = files
        MessageDialog.__init__(self, parent, "Move files to another location?")

    def action_yes(self):
        dlg = wx.DirDialog(self, "Choose directory", "", wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        if dlg.ShowModal() == wx.ID_OK:
            move_files_to_new_path(self.files, dlg.GetPath())
        dlg.Destroy()

    def action_no(self):
        DeleteFiles(self.parent, self.files)


class DeleteFiles(MessageDialog):
    def __init__(self, parent, files):
        self.files = files
        MessageDialog.__init__(self, parent, "Delete associated files?")

    def action_yes(self):
        for f in self.files:
            delete_file(f)


class DeleteFilesFromQuery(MessageDialog):
    def __init__(self, parent, dicom_file_query):
        self.dicom_file_query = dicom_file_query
        MessageDialog.__init__(self, parent, "Delete associated files?")

    def action_yes(self):
        delete_imported_dicom_files(self.dicom_file_query)

    def action_no(self):
        MoveFilesFromQuery(self.parent, self.dicom_file_query)


class MoveFilesFromQuery(MessageDialog):
    def __init__(self, parent, dicom_file_query):
        self.dicom_file_query = dicom_file_query
        MessageDialog.__init__(self, parent, "Move associated files?")

    def action_yes(self):
        dlg = wx.DirDialog(self.parent, "Choose directory", "", wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        if dlg.ShowModal() == wx.ID_OK:
            move_imported_dicom_files(self.dicom_file_query, dlg.GetPath())
        dlg.Destroy()


class RebuildDB(MessageDialog):
    def __init__(self, parent, roi_map, options):
        self.roi_map = roi_map
        self.options = options
        MessageDialog.__init__(self, parent, "Rebuild Database from DICOM")

    def action_yes(self):
        with DVH_SQL() as cnx:
            cnx.reinitialize_database()

        ImportDicomFrame(self.roi_map, self.options, inbox=self.options.IMPORTED_DIR, auto_parse=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.mlc_analyzer.py
"""
Tools for analyzing beam and control point information from DICOM files
Hierarchy of classes:
    Plan -> FxGroup -> Beam -> ControlPoint
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from dicompylercore import dicomparser
import numpy as np
from shapely.geometry import Polygon
from shapely import speedups
from dvha.tools.utilities import flatten_list_of_lists as flatten
from dvha.options import Options


options = Options()

# Enable shapely calculations using C, as opposed to the C++ default
if speedups.available:
    speedups.enable()


class Plan:
    """
    Collect plan information from an RT Plan DICOM file.
    Automatically parses fraction data with FxGroup class
    """
    def __init__(self, rt_plan_file):
        """
        :param rt_plan_file: absolute file path of a DICOM RT Plan file
        """
        rt_plan = dicomparser.read_file(rt_plan_file)

        beam_seq = rt_plan.BeamSequence
        fx_grp_seq = rt_plan.FractionGroupSequence
        self.fx_group = [FxGroup(fx_grp, beam_seq) for fx_grp in fx_grp_seq]

        self.plan_name = rt_plan.RTPlanLabel
        self.patient_name = rt_plan.PatientName
        self.patient_id = rt_plan.PatientID
        self.study_instance_uid = rt_plan.StudyInstanceUID
        self.tps = '%s %s' % (rt_plan.Manufacturer, rt_plan.ManufacturerModelName)

        self.complexity_scores = [fx_grp.complexity_score for fx_grp in self.fx_group]

    def __str__(self):
        summary = ['Patient Name:       %s' % self.patient_name,
                   'Patient MRN:        %s' % self.patient_id,
                   'Study Instance UID: %s' % self.study_instance_uid,
                   'TPS:                %s' % self.tps,
                   'Plan name:          %s' % self.plan_name,
                   '# of Fx Groups:     %s' % len(self.fx_group)]
        return '\n'.join(summary)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        for i, fx_group in enumerate(self.fx_group):
            if not fx_group == other.fx_group[i]:
                return False
        return True


class FxGroup:
    """
    Collect fraction group information from fraction group and beam sequences of a pydicom RT Plan dataset
    Automatically parses beam data with Beam class
    """
    def __init__(self, fx_grp_seq, plan_beam_sequences):
        """
        :param fx_grp_seq: fraction group sequence object
        :type fx_grp_seq: Sequence
        :param plan_beam_sequences: beam sequence object
        :type plan_beam_sequences: Sequence
        """
        self.fxs = fx_grp_seq.NumberOfFractionsPlanned

        meter_set = {}
        for ref_beam in fx_grp_seq.ReferencedBeamSequence:
            ref_beam_num = str(ref_beam.ReferencedBeamNumber)
            meter_set[ref_beam_num] = float(ref_beam.BeamMeterset)

        self.beam = []
        for beam_seq in plan_beam_sequences:
            beam_num = str(beam_seq.BeamNumber)
            if beam_num in meter_set:
                self.beam.append(Beam(beam_seq, meter_set[beam_num]))
        self.beam_count = len(self.beam)
        self.beam_names = [b.name for b in self.beam]

        self.beam = update_missing_jaws(self.beam)

        self.complexity_score = np.sum(np.array([np.sum(beam.complexity_scores) for beam in self.beam]))

    def __eq__(self, other):
        for i, beam in enumerate(self.beam):
            status_str = 'beam %s\n\t%%s with other beam %s' % (self.beam_names[i], other.beam_names[i])
            if not beam == other.beam[i]:
                print(status_str % 'failed')
                return False
            else:
                print(status_str % 'passed')
        return True


class Beam:
    """
    Collect beam information from a beam in a beam sequence of a pydicom RT Plan dataset
    Automatically parses control point data with ControlPoint class
    """
    def __init__(self, beam_dataset, meter_set, ignore_zero_mu_cp=False):
        """
        :param beam_dataset: a pydicom beam sequence object
        :type beam_dataset: Dataset
        :param meter_set: the monitor units for the beam_dataset
        :type meter_set: float
        :param ignore_zero_mu_cp: If True, skip over zero MU control points (e.g., as in Step-N-Shoot beams)
        :type ignore_zero_mu_cp: bool
        """

        cp_seq = beam_dataset.ControlPointSequence
        self.control_point = [ControlPoint(cp) for cp in cp_seq]
        self.control_point_count = len(self.control_point)

        for bld_seq in beam_dataset.BeamLimitingDeviceSequence:
            if hasattr(bld_seq, 'LeafPositionBoundaries'):
                self.leaf_boundaries = bld_seq.LeafPositionBoundaries

        self.jaws = [get_jaws(cp) for cp in self.control_point]
        self.aperture = [get_shapely_from_cp(cp, self.leaf_boundaries) for cp in self.control_point]
        self.mlc_borders = [get_mlc_borders(cp, self.leaf_boundaries) for cp in self.control_point]

        self.gantry_angle = [float(cp.GantryAngle) for cp in cp_seq if hasattr(cp, 'GantryAngle')]
        self.collimator_angle = [float(cp.BeamLimitingDeviceAngle) for cp in cp_seq if hasattr(cp, 'BeamLimitingDeviceAngle')]
        self.couch_angle = [float(cp.PatientSupportAngle) for cp in cp_seq if hasattr(cp, 'PatientSupportAngle')]

        self.meter_set = meter_set
        self.control_point_meter_set = np.append([0], np.diff(np.array([cp.cum_mu for cp in self.control_point])))

        if hasattr(beam_dataset, 'BeamDescription'):
            self.name = beam_dataset.BeamDescription
        else:
            self.name = beam_dataset.BeamName

        cum_mu = [cp.cum_mu * self.meter_set for cp in self.control_point]
        cp_mu = np.diff(np.array(cum_mu)).tolist() + [0]
        x_paths = np.array([get_xy_path_lengths(cp)[0] for cp in self.aperture])
        y_paths = np.array([get_xy_path_lengths(cp)[1] for cp in self.aperture])
        area = [cp.area for cp in self.aperture]
        c1, c2 = options.COMPLEXITY_SCORE_X_WEIGHT, options.COMPLEXITY_SCORE_Y_WEIGHT
        self.complexity_scores = np.divide(np.multiply(np.add(c1*x_paths, c2*y_paths), cp_mu), area) / self.meter_set
        # Complexity score based on:
        # Younge KC, Matuszak MM, Moran JM, McShan DL, Fraass BA, Roberts DA. Penalization of aperture
        # complexity in inversely planned volumetric modulated arc therapy. Med Phys. 2012;39(11):7160–70.

        self.summary = {'cp': range(1, len(self.control_point)+1),
                        'cum_mu_frac': [cp.cum_mu for cp in self.control_point],
                        'cum_mu': cum_mu,
                        'cp_mu': cp_mu,
                        'gantry': self.gantry_angle,
                        'collimator': self.collimator_angle,
                        'couch': self.couch_angle,
                        'jaw_x1': [j['x_min']/10 for j in self.jaws],
                        'jaw_x2': [j['x_max']/10 for j in self.jaws],
                        'jaw_y1': [j['y_min']/10 for j in self.jaws],
                        'jaw_y2': [j['y_max']/10 for j in self.jaws],
                        'area': np.divide(area, 100.).tolist(),
                        'x_perim': np.divide(x_paths, 10.).tolist(),
                        'y_perim': np.divide(y_paths, 10.).tolist(),
                        'perim': np.divide(np.add(x_paths, y_paths), 10.).tolist(),
                        'cmp_score': self.complexity_scores.tolist()}

        for key in self.summary:
            if len(self.summary[key]) == 1:
                self.summary[key] = self.summary[key] * len(self.summary['cp'])

        if ignore_zero_mu_cp:
            non_zero_indices = [i for i, value in enumerate(self.summary['cp_mu']) if value != 0]
            for key in list(self.summary):
                self.summary[key] = [self.summary[key][i] for i in non_zero_indices]

    def __eq__(self, other):
        for i, cp in enumerate(self.control_point):
            if not cp == other.control_point[i]:
                print('cp %s failed' % i)
                return False
        return True


class ControlPoint:
    """
    Collect control point information from a ControlPointSequence in a beam dataset of a pydicom RT Plan dataset
    """
    def __init__(self, cp_seq):
        """
        :param cp_seq: control point sequence object
        :type cp_seq: Sequence
        """
        cp = {'cum_mu': float(cp_seq.CumulativeMetersetWeight)}
        for device_position_seq in cp_seq.BeamLimitingDevicePositionSequence:
            leaf_jaw_type = str(device_position_seq.RTBeamLimitingDeviceType).lower()
            if leaf_jaw_type.startswith('mlc'):
                cp['leaf_type'] = leaf_jaw_type
                leaf_jaw_type = 'mlc'

            positions = np.array(list(map(float, device_position_seq.LeafJawPositions)))
            mid_index = int(len(positions) / 2)
            cp[leaf_jaw_type] = [positions[:mid_index],
                                 positions[mid_index:]]

        if 'leaf_type' not in list(cp):
            cp['leaf_type'] = False

        for key in cp:
            setattr(self, key, cp[key])

    def __eq__(self, other):
        if abs(self.cum_mu - other.cum_mu) > 0.00001:
            return False
        for side in [0, 1]:
            for i, pos in enumerate(self.mlc[side]):
                if abs(pos - other.mlc[side][i]) > 0.0001:
                    print(abs(pos - other.mlc[side][i]))
        return True


def get_mlc_borders(control_point, leaf_boundaries):
    """
    This function returns the boundaries of each MLC leaf for purposes of displaying a beam's eye view using
    bokeh's quad() glyph
    :param control_point: a ControlPoint from a pydicom ControlPoint Sequence
    :type control_point: Dataset
    :param leaf_boundaries: a LeafPositionBoundaries object from the BeamLimitingDeviceSequence
    :type leaf_boundaries: MultiValue
    :return: the boundaries of each leaf within the control point
    :rtype: dict
    """
    top = leaf_boundaries[0:-1] + leaf_boundaries[0:-1]
    top = [float(i) for i in top]
    bottom = leaf_boundaries[1::] + leaf_boundaries[1::]
    bottom = [float(i) for i in bottom]
    left = [- options.MAX_FIELD_SIZE_X / 2] * len(control_point.mlc[0])
    left.extend(control_point.mlc[1])
    right = control_point.mlc[0].tolist()
    right.extend([options.MAX_FIELD_SIZE_X / 2] * len(control_point.mlc[1]))

    return {'top': top,
            'bottom': bottom,
            'left': left,
            'right': right}


def get_shapely_from_cp(control_point, leaf_boundaries):
    """
    This function will return the outline of MLCs within jaws
    :param control_point: a ControlPoint from a pydicom ControlPoint Sequence
    :type control_point: Dataset
    :param leaf_boundaries: a LeafPositionBoundaries object from the BeamLimitingDeviceSequence
    :type leaf_boundaries: MultiValue
    :return: a shapely object of the complete MLC aperture as one shape (including MLC overlap)
    :rtype: Polygon
    """
    lb = leaf_boundaries
    mlc = control_point.mlc
    jaws = get_jaws(control_point)
    x_min, x_max = jaws['x_min'], jaws['x_max']
    y_min, y_max = jaws['y_min'], jaws['y_max']

    jaw_points = [(x_min, y_min), (x_min, y_max), (x_max, y_max), (x_max, y_min)]
    jaw_shapely = Polygon(jaw_points)

    if control_point.leaf_type == 'mlcx':
        a = flatten([[(m, lb[i]), (m, lb[i+1])] for i, m in enumerate(mlc[0])])
        b = flatten([[(m, lb[i]), (m, lb[i+1])] for i, m in enumerate(mlc[1])])
    elif control_point.leaf_type == 'mlcy':
        a = flatten([[(lb[i], m), (lb[i + 1], m)] for i, m in enumerate(mlc[0])])
        b = flatten([[(lb[i], m), (lb[i + 1], m)] for i, m in enumerate(mlc[1])])
    else:
        return jaw_shapely

    mlc_points = a + b[::-1]  # concatenate a and reverse(b)
    mlc_aperture = Polygon(mlc_points).buffer(0)

    # This function is very slow, since jaws are rectangular, perhaps there's a more efficient method?
    aperture = mlc_aperture.intersection(jaw_shapely)

    return aperture


def get_jaws(control_point):
    """
    Get the jaw positions of a control point
    :param control_point: a ControlPoint from a pydicom ControlPoint Sequence
    :type control_point: Dataset
    :return: jaw positions (or max field size in lieu of a jaw)
    :rtype: dict
    """

    cp = control_point

    # Determine jaw opening
    if hasattr(cp, 'asymy'):
        y_min = min(cp.asymy)
        y_max = max(cp.asymy)
    else:
        y_min = -options.MAX_FIELD_SIZE_Y / 2.
        y_max = options.MAX_FIELD_SIZE_Y / 2.
    if hasattr(cp, 'asymx'):
        x_min = min(cp.asymx)
        x_max = max(cp.asymx)
    else:
        x_min = -options.MAX_FIELD_SIZE_X / 2.
        x_max = options.MAX_FIELD_SIZE_X / 2.

    jaws = {'x_min': float(x_min),
            'x_max': float(x_max),
            'y_min': float(y_min),
            'y_max': float(y_max)}

    return jaws


def get_xy_path_lengths(shapely_object):
    """
    Get the x and y path lengths of a a Shapely object
    :param shapely_object: either 'GeometryCollection', 'MultiPolygon', or 'Polygon'
    :return: path lengths in the x and y directions
    :rtype: list
    """
    path = np.array([0., 0.])
    if shapely_object.type == 'GeometryCollection':
        for geometry in shapely_object.geoms:
            if geometry.type in {'MultiPolygon', 'Polygon'}:
                path = np.add(path, get_xy_path_lengths(geometry))
    elif shapely_object.type == 'MultiPolygon':
        for shape in shapely_object:
            path = np.add(path, get_xy_path_lengths(shape))
    elif shapely_object.type == 'Polygon':
        x, y = np.array(shapely_object.exterior.xy[0]), np.array(shapely_object.exterior.xy[1])
        path = np.array([np.sum(np.abs(np.diff(x))), np.sum(np.abs(np.diff(y)))])

    return path.tolist()


def update_missing_jaws(beam_list):
    """
    In plans with static jaws throughout the beam, jaw positions may not be found in each control point
    :param beam_list: a list of Beam Class objects
    :return: a list of Beam Class objects, but any control point where jaws are set to max will be replaced by the
    first control point jaw settings
    :rtype: list
    """

    for i, beam in enumerate(beam_list):
        for j, cp in enumerate(beam.jaws):
            if cp['x_min'] == -options.MAX_FIELD_SIZE_X / 2 and \
                            cp['x_max'] == options.MAX_FIELD_SIZE_X / 2 and \
                            cp['y_min'] == -options.MAX_FIELD_SIZE_Y / 2 and \
                            cp['y_max'] == options.MAX_FIELD_SIZE_Y / 2:
                beam_list[i].jaws[j] = beam.jaws[0]

    return beam_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dvha_app.py
"""
Script to start DVH Analytics
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics


import dvha.main
import multiprocessing

if __name__ == "__main__":
    multiprocessing.freeze_support()
    dvha.main.start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.roi_map.py
"""
Classes for viewing and editing the roi map, and updating the database with changes
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
import wx.html2
from datetime import datetime
from threading import Thread
from pubsub import pub
from dvha.db.sql_connector import DVH_SQL, echo_sql_db
from dvha.dialogs.roi_map import AddPhysician, AddPhysicianROI, AddVariation, MoveVariationDialog,\
    RenamePhysicianDialog, RenamePhysicianROIDialog, RenameInstitutionalROIDialog, LinkPhysicianROI
from dvha.models.data_table import DataTable
from dvha.models.plot import PlotROIMap
from dvha.tools.utilities import get_selected_listctrl_items, MessageDialog, get_elapsed_time, get_window_size,\
    set_frame_icon, set_msw_background_color, is_windows
from dvha.tools.roi_map_generator import ROIMapGenerator
from dvha.tools.roi_name_manager import clean_name


class ROIMapFrame(wx.Frame):
    """
    Class to view and edit roi map
    """
    def __init__(self, roi_map):
        """
        :param roi_map: roi map object
        :type roi_map: DatabaseROIs
        """
        wx.Frame.__init__(self, None, title='ROI Map')
        set_frame_icon(self)

        self.roi_map = roi_map

        self.window_size = get_window_size(0.893, 0.762)
        self.SetSize(self.window_size)
        self.window = wx.SplitterWindow(self, wx.ID_ANY)
        self.window_tree = wx.Panel(self.window, wx.ID_ANY, style=wx.BORDER_SUNKEN)

        self.combo_box_tree_plot_data = wx.ComboBox(self.window_tree, wx.ID_ANY,
                                                    choices=['All', 'Linked', 'Unlinked', 'Branched'],
                                                    style=wx.CB_DROPDOWN | wx.CB_READONLY)

        self.plot = PlotROIMap(self.window_tree, roi_map)
        self.window_editor = wx.Panel(self.window, wx.ID_ANY, style=wx.BORDER_SUNKEN)

        self.combo_box_physician = wx.ComboBox(self.window_editor, wx.ID_ANY, choices=self.roi_map.get_physicians(),
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.combo_box_physician_roi = wx.ComboBox(self.window_editor, wx.ID_ANY, choices=[],
                                                   style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.list_ctrl_variations = wx.ListCtrl(self.window_editor, wx.ID_ANY,
                                                style=wx.LC_NO_HEADER | wx.LC_REPORT | wx.BORDER_SUNKEN)
        self.button_variation_select_all = wx.Button(self.window_editor, wx.ID_ANY, "Select All")
        self.button_variation_deselect_all = wx.Button(self.window_editor, wx.ID_ANY, "Deselect All")
        self.button_variation_add = wx.Button(self.window_editor, wx.ID_ANY, "Add")
        self.button_variation_delete = wx.Button(self.window_editor, wx.ID_ANY, "Delete")
        self.button_variation_move = wx.Button(self.window_editor, wx.ID_ANY, "Move")

        self.button_variation_move.Disable()
        self.button_variation_delete.Disable()
        self.button_variation_deselect_all.Disable()

        self.button_physician = {'add': wx.Button(self.window_editor, wx.ID_ANY, "+"),
                                 'del': wx.Button(self.window_editor, wx.ID_ANY, "-"),
                                 'edit': wx.Button(self.window_editor, wx.ID_ANY, "Δ")}
        self.button_physician_roi = {'add': wx.Button(self.window_editor, wx.ID_ANY, "+"),
                                     'del': wx.Button(self.window_editor, wx.ID_ANY, "-"),
                                     'edit': wx.Button(self.window_editor, wx.ID_ANY, "Δ")}

        self.button_link_physician_roi = wx.Button(self.window_editor, wx.ID_ANY, "Link")
        self.button_link_physician_roi.Disable()

        self.combo_box_uncategorized_ignored = wx.ComboBox(self.window_editor, wx.ID_ANY,
                                                           choices=["Uncategorized", "Ignored"],
                                                           style=wx.CB_DROPDOWN | wx.CB_READONLY)
        self.combo_box_uncategorized_ignored_roi = wx.ComboBox(self.window_editor, wx.ID_ANY, choices=[],
                                                               style=wx.CB_DROPDOWN)
        self.button_uncategorized_ignored_delete = wx.Button(self.window_editor, wx.ID_ANY, "Delete DVH")
        self.button_uncategorized_ignored_ignore = wx.Button(self.window_editor, wx.ID_ANY, "Ignore DVH")
        self.combo_box_physician_roi_merge = {'a': wx.ComboBox(self.window_editor, wx.ID_ANY,
                                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                                              'b': wx.ComboBox(self.window_editor, wx.ID_ANY,
                                                               style=wx.CB_DROPDOWN | wx.CB_READONLY)}
        self.button_merge = wx.Button(self.window_editor, wx.ID_ANY, "Merge")

        self.list_ctrl_tg263 = wx.ListCtrl(self.window_editor, wx.ID_ANY, style=wx.LC_REPORT | wx.BORDER_SUNKEN)
        self.roi_map_gen = ROIMapGenerator()
        self.roi_map_gen.prep_data_for_roi_map_gui()
        self.data_table_tg263 = DataTable(self.list_ctrl_tg263, columns=self.roi_map_gen.keys)
        combo_map = {'anatomy': self.roi_map_gen.anatomic_groups,
                     'target': self.roi_map_gen.target_types,
                     'major': self.roi_map_gen.major_categories,
                     'minor': self.roi_map_gen.minor_categories}
        self.combo_box_tg263 = {key: wx.ComboBox(self.window_editor, wx.ID_ANY,
                                                 choices=['All'] + choices, style=wx.CB_DROPDOWN | wx.CB_READONLY)
                                for key, choices in combo_map.items()}

        self.button_save_and_update = wx.Button(self.window_editor, wx.ID_ANY, "Save and Update Database")
        self.button_cancel = wx.Button(self.window_editor, wx.ID_ANY, "Cancel Changes and Reload")

        self.uncategorized_variations = {}

        self.columns = ['Variations']
        self.data_table = DataTable(self.list_ctrl_variations, columns=self.columns, widths=[490])

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.plot.update_roi_map_source_data(self.physician)

        self.run()

    def __set_properties(self):
        self.combo_box_uncategorized_ignored.SetSelection(0)
        self.button_uncategorized_ignored_ignore.SetMinSize((110, 20))

        self.combo_box_physician.SetValue('DEFAULT')
        self.update_physician_rois()
        self.update_variations()

        self.combo_box_physician.SetValue('DEFAULT')
        self.combo_box_tree_plot_data.SetValue('ALL')

        self.update_uncategorized_ignored_choices()

        self.window_tree.SetBackgroundColour('white')

        for button in self.button_physician.values():
            button.SetMaxSize((25, 25))
        for button in self.button_physician_roi.values():
            button.SetMaxSize((25, 25))

        self.update_physician_enable()
        self.update_merge_physician_rois()

        if is_windows():  # combo_boxes here display a doubled bottom border on MSW
            combo_boxes = [self.combo_box_physician, self.combo_box_physician_roi,
                           self.combo_box_uncategorized_ignored, self.combo_box_uncategorized_ignored_roi,
                           self.combo_box_physician_roi_merge['a'], self.combo_box_physician_roi_merge['b']]
            for combo_box in combo_boxes:
                combo_box.SetMinSize((combo_box.GetSize()[0], 26))
            self.button_uncategorized_ignored_ignore.SetMinSize((self.button_uncategorized_ignored_ignore.GetSize()[0],
                                                                 self.button_uncategorized_ignored_delete.GetSize()[1]))

        self.data_table_tg263.set_data(self.roi_map_gen.tg_263, columns=self.roi_map_gen.keys)
        self.data_table_tg263.set_column_widths(auto=True)
        for combo_box in self.combo_box_tg263.values():
            combo_box.SetValue('All')

        # These combo_boxes get a height of 30 since adding TG263 combo_boxes?
        for combo_box in self.combo_box_physician_roi_merge.values():
            combo_box.SetMaxSize((1000, 26))

    def __do_bind(self):
        self.window_tree.Bind(wx.EVT_COMBOBOX, self.on_plot_data_type_change, id=self.combo_box_tree_plot_data.GetId())

        self.window_editor.Bind(wx.EVT_COMBOBOX, self.update_uncategorized_ignored_choices,
                                id=self.combo_box_uncategorized_ignored.GetId())

        self.window_editor.Bind(wx.EVT_COMBOBOX, self.physician_ticker, id=self.combo_box_physician.GetId())
        self.window_editor.Bind(wx.EVT_COMBOBOX, self.physician_roi_ticker, id=self.combo_box_physician_roi.GetId())
        self.window_editor.Bind(wx.EVT_COMBOBOX, self.uncategorized_ticker, id=self.combo_box_uncategorized_ignored.GetId())

        self.window_editor.Bind(wx.EVT_BUTTON, self.add_physician, id=self.button_physician['add'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_delete_physician, id=self.button_physician['del'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_edit_physician, id=self.button_physician['edit'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_link_physician_roi, id=self.button_link_physician_roi.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.add_physician_roi, id=self.button_physician_roi['add'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_delete_physician_roi, id=self.button_physician_roi['del'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_edit_physician_roi, id=self.button_physician_roi['edit'].GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.select_all_variations, id=self.button_variation_select_all.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.deselect_all_variations, id=self.button_variation_deselect_all.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.add_variation, id=self.button_variation_add.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.move_variations, id=self.button_variation_move.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.delete_variations, id=self.button_variation_delete.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_delete_dvh, id=self.button_uncategorized_ignored_delete.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_ignore_dvh, id=self.button_uncategorized_ignored_ignore.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_merge, id=self.button_merge.GetId())

        for combo_box in self.combo_box_tg263.values():
            self.window_editor.Bind(wx.EVT_COMBOBOX, self.update_tg263_table, id=combo_box.GetId())
        self.window_editor.Bind(wx.EVT_LIST_COL_CLICK, self.sort_tg263_table, self.list_ctrl_tg263)

        self.window_editor.Bind(wx.EVT_BUTTON, self.save_and_update, id=self.button_save_and_update.GetId())
        self.window_editor.Bind(wx.EVT_BUTTON, self.on_cancel, id=self.button_cancel.GetId())
        self.Bind(wx.EVT_CLOSE, self.on_close)

        self.window_editor.Bind(wx.EVT_COMBOBOX, self.update_merge_enable,
                                id=self.combo_box_physician_roi_merge['a'].GetId())
        self.window_editor.Bind(wx.EVT_COMBOBOX, self.update_merge_enable,
                                id=self.combo_box_physician_roi_merge['b'].GetId())
        self.window_editor.Bind(wx.EVT_LIST_ITEM_SELECTED, self.update_button_variation_enable,
                                id=self.list_ctrl_variations.GetId())
        self.window_editor.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.update_button_variation_enable,
                                id=self.list_ctrl_variations.GetId())

    def __do_layout(self):

        sizer_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_editor = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_roi_merger = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician_roi_merger_merge = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_roi_b = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_roi_a = wx.BoxSizer(wx.VERTICAL)
        sizer_uncategorized_ignored = wx.StaticBoxSizer(
            wx.StaticBox(self.window_editor, wx.ID_ANY, "Uncategorized / Ignored"), wx.HORIZONTAL)
        sizer_uncategorized_ignored_ignore = wx.BoxSizer(wx.VERTICAL)
        sizer_uncategorized_ignored_delete = wx.BoxSizer(wx.VERTICAL)
        sizer_uncategorized_ignored_roi = wx.BoxSizer(wx.VERTICAL)
        sizer_uncategorized_ignored_type = wx.BoxSizer(wx.VERTICAL)
        sizer_tree = wx.BoxSizer(wx.VERTICAL)
        sizer_tree_input = wx.BoxSizer(wx.HORIZONTAL)
        sizer_tree_plot_data = wx.BoxSizer(wx.VERTICAL)
        sizer_roi_manager = wx.BoxSizer(wx.VERTICAL)
        sizer_variation_buttons = wx.BoxSizer(wx.VERTICAL)
        sizer_variation_table = wx.BoxSizer(wx.VERTICAL)
        sizer_map_editor = wx.StaticBoxSizer(wx.StaticBox(self.window_editor, wx.ID_ANY, "ROI Map Editor"), wx.VERTICAL)
        sizer_variations = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician_roi = wx.BoxSizer(wx.VERTICAL)
        sizer_physician = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_row = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician_roi_row = wx.BoxSizer(wx.HORIZONTAL)
        sizer_tg263 = wx.StaticBoxSizer(wx.StaticBox(self.window_editor, wx.ID_ANY, "TG-263 (for reference)"),
                                        wx.VERTICAL)
        sizer_tg263_filters = wx.BoxSizer(wx.HORIZONTAL)
        sizer_tg263_col = {key: wx.BoxSizer(wx.VERTICAL) for key in list(self.combo_box_tg263)}
        sizer_tg263_table = wx.BoxSizer(wx.VERTICAL)
        sizer_save_cancel_buttons = wx.BoxSizer(wx.HORIZONTAL)

        label_physician = wx.StaticText(self.window_editor, wx.ID_ANY, "Physician:")
        sizer_physician.Add(label_physician, 0, 0, 0)
        sizer_physician_row.Add(self.combo_box_physician, 1, wx.EXPAND | wx.RIGHT, 5)
        sizer_physician_row.Add(self.button_physician['add'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_physician_row.Add(self.button_physician['del'], 0, wx.RIGHT, 5)
        sizer_physician_row.Add(self.button_physician['edit'], 0, wx.RIGHT, 10)
        sizer_physician.Add(sizer_physician_row, 1, wx.EXPAND, 0)

        self.label_physician_roi = wx.StaticText(self.window_editor, wx.ID_ANY, "Institutional ROI:")
        sizer_physician_roi.Add(self.label_physician_roi, 0, 0, 0)
        sizer_physician_roi_row.Add(self.combo_box_physician_roi, 1, wx.EXPAND | wx.RIGHT, 5)
        sizer_physician_roi_row.Add(self.button_link_physician_roi, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_physician_roi_row.Add(self.button_physician_roi['add'], 0, wx.LEFT | wx.RIGHT, 5)
        sizer_physician_roi_row.Add(self.button_physician_roi['del'], 0, wx.RIGHT, 5)
        sizer_physician_roi_row.Add(self.button_physician_roi['edit'], 0, wx.RIGHT, 10)
        sizer_physician_roi.Add(sizer_physician_roi_row, 0, wx.EXPAND, 0)

        sizer_map_editor.Add(sizer_physician, 0, wx.ALL | wx.EXPAND, 5)
        sizer_map_editor.Add(sizer_physician_roi, 0, wx.ALL | wx.EXPAND, 5)

        label_variations = wx.StaticText(self.window_editor, wx.ID_ANY, "Variations:")
        label_variations_buttons = wx.StaticText(self.window_editor, wx.ID_ANY, " ")
        sizer_variation_table.Add(label_variations, 0, 0, 0)
        sizer_variation_table.Add(self.list_ctrl_variations, 1, wx.BOTTOM | wx.EXPAND, 15)
        sizer_variations.Add(sizer_variation_table, 1, wx.EXPAND | wx.ALL, 5)
        sizer_variation_buttons.Add(label_variations_buttons, 0, 0, 0)
        sizer_variation_buttons.Add(self.button_variation_add, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5)
        sizer_variation_buttons.Add(self.button_variation_delete, 0, wx.EXPAND | wx.ALL, 5)
        sizer_variation_buttons.Add(self.button_variation_move, 0, wx.EXPAND | wx.ALL, 5)
        sizer_variation_buttons.Add(self.button_variation_select_all, 0, wx.EXPAND | wx.ALL, 5)
        sizer_variation_buttons.Add(self.button_variation_deselect_all, 0, wx.EXPAND | wx.ALL, 5)
        sizer_variations.Add(sizer_variation_buttons, 0, wx.EXPAND, 0)

        sizer_map_editor.Add(sizer_variations, 0, wx.EXPAND, 0)

        label_physician_roi_a = wx.StaticText(self.window_editor, wx.ID_ANY, "Merge Physician ROI A:")
        sizer_physician_roi_a.Add(label_physician_roi_a, 0, wx.EXPAND | wx.LEFT, 5)
        sizer_physician_roi_a.Add(self.combo_box_physician_roi_merge['a'], 1,
                                  wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_physician_roi_merger.Add(sizer_physician_roi_a, 1, wx.EXPAND, 0)
        label_physician_roi_b = wx.StaticText(self.window_editor, wx.ID_ANY, "Into Physician ROI B:")
        sizer_physician_roi_b.Add(label_physician_roi_b, 0, wx.EXPAND, 0)
        sizer_physician_roi_b.Add(self.combo_box_physician_roi_merge['b'], 1,
                                  wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_physician_roi_merger.Add(sizer_physician_roi_b, 1, wx.EXPAND, 0)
        sizer_physician_roi_merger_merge.Add((20, 16), 0, 0, 0)
        sizer_physician_roi_merger_merge.Add(self.button_merge, 0, wx.ALL, 5)
        sizer_physician_roi_merger.Add(sizer_physician_roi_merger_merge, 0, wx.ALL | wx.EXPAND, 0)
        sizer_map_editor.Add(sizer_physician_roi_merger, 0, wx.EXPAND, 0)

        sizer_save_cancel_buttons.Add(self.button_save_and_update, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 40)
        sizer_save_cancel_buttons.Add(self.button_cancel, 1, wx.EXPAND | wx.LEFT| wx.RIGHT, 40)
        sizer_map_editor.Add((10, 10), 0, 0, 0)
        sizer_map_editor.Add(sizer_save_cancel_buttons, 0, wx.EXPAND | wx.ALL, 5)

        sizer_roi_manager.Add(sizer_map_editor, 1, wx.EXPAND, 0)
        sizer_editor.Add(sizer_roi_manager, 0, wx.EXPAND | wx.ALL, 5)

        label_tree_plot_data = wx.StaticText(self.window_tree, wx.ID_ANY, 'Institutional Data to Display:')
        sizer_tree_plot_data.Add(label_tree_plot_data, 0, wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_tree_plot_data.Add(self.combo_box_tree_plot_data, 0, wx.EXPAND | wx.LEFT | wx.BOTTOM | wx.RIGHT, 5)

        sizer_tree_input.Add(sizer_tree_plot_data, 0, wx.EXPAND | wx.LEFT | wx.TOP, 5)

        sizer_tree.Add(sizer_tree_input, 0, wx.EXPAND, 0)
        sizer_tree.Add(self.plot.layout, 1, wx.EXPAND, 0)
        self.window_tree.SetSizer(sizer_tree)

        label_uncategorized_ignored = wx.StaticText(self.window_editor, wx.ID_ANY, "Type:")
        label_uncategorized_ignored.SetMinSize((38, 16))
        sizer_uncategorized_ignored_type.Add(label_uncategorized_ignored, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_uncategorized_ignored_type.Add(self.combo_box_uncategorized_ignored, 1, wx.EXPAND | wx.BOTTOM | wx.LEFT | wx.RIGHT, 5)
        sizer_uncategorized_ignored.Add(sizer_uncategorized_ignored_type, 1, wx.EXPAND, 0)
        label_uncategorized_ignored_roi = wx.StaticText(self.window_editor, wx.ID_ANY, "ROI:")
        label_uncategorized_ignored_roi.SetMinSize((30, 16))
        sizer_uncategorized_ignored_roi.Add(label_uncategorized_ignored_roi, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_uncategorized_ignored_roi.Add(self.combo_box_uncategorized_ignored_roi, 1, wx.EXPAND | wx.BOTTOM | wx.LEFT | wx.RIGHT, 5)
        sizer_uncategorized_ignored.Add(sizer_uncategorized_ignored_roi, 1, wx.EXPAND, 0)
        sizer_uncategorized_ignored_delete.Add((20, 16), 0, 0, 0)
        sizer_uncategorized_ignored_delete.Add(self.button_uncategorized_ignored_delete, 0, wx.ALL, 5)
        sizer_uncategorized_ignored.Add(sizer_uncategorized_ignored_delete, 0, wx.EXPAND, 0)
        sizer_uncategorized_ignored_ignore.Add((20, 16), 0, 0, 0)
        sizer_uncategorized_ignored_ignore.Add(self.button_uncategorized_ignored_ignore, 0, wx.ALL, 5)
        sizer_uncategorized_ignored.Add(sizer_uncategorized_ignored_ignore, 0, wx.EXPAND, 0)
        sizer_editor.Add(sizer_uncategorized_ignored, 0, wx.ALL | wx.EXPAND, 5)

        sizer_tg263_table.Add(self.list_ctrl_tg263, 1, wx.EXPAND, 0)

        label_tg263 = {key: wx.StaticText(self.window_editor, wx.ID_ANY, key.capitalize() + ':')
                       for key in list(self.combo_box_tg263)}
        for key in ['major', 'minor', 'anatomy', 'target']:
            sizer_tg263_col[key].Add(label_tg263[key])
            sizer_tg263_col[key].Add(self.combo_box_tg263[key], 0, wx.EXPAND, 0)
            sizer_tg263_filters.Add(sizer_tg263_col[key], 0, wx.EXPAND, 0)

        sizer_tg263.Add(sizer_tg263_filters, 0, 0, 0)
        sizer_tg263.Add(sizer_tg263_table, 1, wx.EXPAND, 0)
        sizer_editor.Add(sizer_tg263, 1, wx.EXPAND | wx.ALL, 5)

        self.window_editor.SetSizer(sizer_editor)
        self.window.SplitVertically(self.window_tree, self.window_editor)
        self.window.SetSashPosition(int(self.window_size[0] * 0.55))
        sizer_wrapper.Add(self.window, 1, wx.EXPAND, 0)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Centre()

    def run(self):
        self.Show()

    @staticmethod
    def update_combo_box_choices(combo_box, choices, value):
        if not value:
            value = combo_box.GetValue()
        combo_box.Clear()
        combo_box.AppendItems(sorted(choices))
        combo_box.SetValue(value)

    def update_roi_map(self):
        self.plot.update_roi_map_source_data(self.physician, plot_type=self.plot_data_type,
                                             y_shift=self.combo_box_physician_roi.GetValue())

    @property
    def physician(self):
        return self.combo_box_physician.GetValue()

    @property
    def physician_roi(self):
        return self.combo_box_physician_roi.GetValue()

    @property
    def plot_data_type(self):
        return self.combo_box_tree_plot_data.GetValue()

    def physician_ticker(self, evt):
        self.update_physician_roi_label()
        self.update_physician_enable()
        self.update_all(skip_physicians=True)

    def on_plot_data_type_change(self, evt):
        self.update_roi_map()

    def update_uncategorized_ignored_choices(self, *args):
        ignored_variations = self.combo_box_uncategorized_ignored.GetValue() == 'Ignored'
        self.uncategorized_variations = self.get_uncategorized_variations(self.physician,
                                                                          ignored_variations=ignored_variations)
        choices = list(self.uncategorized_variations)
        choices.sort()
        if not choices:
            choices = ['None']
            self.button_uncategorized_ignored_delete.Disable()
            self.button_uncategorized_ignored_ignore.Disable()
        else:
            self.button_uncategorized_ignored_delete.Enable()
            self.button_uncategorized_ignored_ignore.Enable()
        self.combo_box_uncategorized_ignored_roi.Clear()
        self.combo_box_uncategorized_ignored_roi.Append(choices)
        self.combo_box_uncategorized_ignored_roi.SetValue(choices[0])

    @staticmethod
    def get_uncategorized_variations(physician, ignored_variations=False):
        if echo_sql_db():
            with DVH_SQL() as cnx:
                physician = clean_name(physician).upper()
                condition = "physician_roi = '%s'" % ['uncategorized', 'ignored'][ignored_variations]
                cursor_rtn = cnx.query('dvhs', 'roi_name, study_instance_uid', condition)
                new_variations = {}
                for row in cursor_rtn:
                    variation = str(row[0])
                    study_instance_uid = str(row[1])
                    physician_db = cnx.get_unique_values('Plans', 'physician',
                                                         "study_instance_uid = '%s'" % study_instance_uid)
                    if physician_db and physician_db[0] == physician:
                        if variation not in list(new_variations):
                            new_variations[variation] = {'roi_name': variation,
                                                         'study_instance_uid': [study_instance_uid]}
                        else:
                            new_variations[variation]['study_instance_uid'].append(study_instance_uid)
                return new_variations

    @property
    def selected_indices(self):
        return get_selected_listctrl_items(self.list_ctrl_variations)

    def update_button_variation_enable(self, *args):
        if self.selected_indices:
            self.button_variation_move.Enable()
            self.button_variation_delete.Enable()
            self.button_variation_deselect_all.Enable()
        else:
            self.button_variation_move.Disable()
            self.button_variation_delete.Disable()
            self.button_variation_deselect_all.Disable()

        self.button_variation_select_all.Enable(self.variation_count > 0)

    def update_variations(self):
        self.data_table.set_data(self.variation_table_data, self.columns)
        self.update_button_variation_enable()

    def physician_roi_ticker(self, evt):
        self.update_variations()
        self.update_roi_map()

    def update_physicians(self, old_physicians=None):

        choices = list(self.roi_map.physicians)
        new = choices[0]
        if old_physicians:
            new = list(set(choices) - set(old_physicians))
            if new:
                new = clean_name(new[0]).upper()

        self.update_combo_box_choices(self.combo_box_physician, choices, new)
        self.update_physician_roi_label()
        self.update_physician_enable()

    def update_physician_rois(self, old_physician_rois=None):
        choices = self.roi_map.get_physician_rois(self.physician)
        new = choices[0]
        if old_physician_rois:
            new = list(set(choices) - set(old_physician_rois))
            if new:
                new = new[0]
        self.update_combo_box_choices(self.combo_box_physician_roi, choices, new)

    @property
    def variations(self):
        variations = self.roi_map.get_variations(self.physician, self.physician_roi)
        variations = list(set(variations) - {self.physician_roi})  # remove physician roi
        variations.sort()
        return variations

    @property
    def variation_table_data(self):
        return {'Variations': self.variations}

    def add_physician_roi(self, evt):
        old_physician_rois = self.roi_map.get_physician_rois(self.physician)
        dlg = AddPhysicianROI(self, self.physician, self.roi_map, institutional_mode=self.physician == 'DEFAULT')
        if dlg.res == wx.ID_OK:
            self.update_all(old_physician_rois=old_physician_rois, skip_physicians=True)

    def add_physician(self, evt):
        old_physicians = list(self.roi_map.physicians)
        dlg = AddPhysician(self.roi_map)
        if dlg.res == wx.ID_OK:
            self.update_all(old_physicians=old_physicians)

    @property
    def variation_count(self):
        return len(self.variations)

    @property
    def selected_values(self):
        return [self.list_ctrl_variations.GetItem(i, 0).GetText() for i in self.selected_indices]

    def select_all_variations(self, evt):
        self.apply_global_selection()

    def deselect_all_variations(self, evt):
        self.apply_global_selection(on=0)

    def apply_global_selection(self, on=1):
        for i in range(self.variation_count):
            self.list_ctrl_variations.Select(i, on=on)

    def delete_variations(self, evt):
        self.roi_map.delete_variations(self.physician, self.physician_roi, self.selected_values)
        self.update_variations()
        self.update_roi_map()

    def add_variation(self, evt):
        AddVariation(self, self.physician, self.roi_map, self.physician_roi)
        self.update_variations()
        self.update_roi_map()

    def move_variations(self, evt):
        choices = [roi for roi in self.roi_map.get_physician_rois(self.physician) if roi != self.physician_roi]
        MoveVariationDialog(self, self.selected_values, self.physician, self.physician_roi, choices, self.roi_map)
        self.update_variations()
        self.update_roi_map()

    def on_delete_physician(self, evt):
        MessageDialog(self, 'Delete Physician %s?' % self.physician, action_yes_func=self.delete_physician)

    def delete_physician(self):
        self.roi_map.delete_physician(self.physician)
        self.update_all()

    def on_delete_physician_roi(self, evt):
        if self.physician == 'DEFAULT':
            MessageDialog(self, "Delete Institutional ROI %s?" % self.physician_roi,
                          action_yes_func=self.delete_institutional_roi)
        else:
            MessageDialog(self, "Delete Physician ROI %s?" % self.physician_roi,
                          action_yes_func=self.delete_physician_roi)

    def delete_physician_roi(self):
        self.roi_map.delete_physician_roi(self.physician, self.physician_roi)
        self.update_all(skip_physicians=True)

    def delete_institutional_roi(self):
        self.roi_map.delete_institutional_roi(self.physician_roi)
        self.update_all(skip_physicians=True)

    def on_delete_dvh(self, evt):
        MessageDialog(self, "Delete all DVHs named %s for %s?" % (self.dvh, self.physician),
                      message="Are you sure? This cannot be undone!",
                      action_yes_func=self.delete_dvh)

    def delete_dvh(self):
        with DVH_SQL() as cnx:
            for uid in self.dvh_uids:
                cnx.delete_dvh(self.dvh, uid)
        self.update_uncategorized_ignored_choices()

    def on_ignore_dvh(self, evt):
        msg_type = ['Unignore', 'Ignore'][self.button_uncategorized_ignored_ignore.GetLabelText() == 'Ignore DVH']
        MessageDialog(self, "%s all DVHs named %s for %s?" % (msg_type, self.dvh, self.physician),
                      action_yes_func=self.ignore_dvh)

    def ignore_dvh(self):
        unignore = self.button_uncategorized_ignored_ignore.GetLabelText() == 'Unignore DVH'
        with DVH_SQL() as cnx:
            for uid in self.dvh_uids:
                cnx.ignore_dvh(self.dvh, uid, unignore=unignore)
        self.update_uncategorized_ignored_choices()

    @property
    def dvh(self):
        return self.combo_box_uncategorized_ignored_roi.GetValue()

    @property
    def dvh_uids(self):
        return self.uncategorized_variations[self.dvh]['study_instance_uid']

    def update_all(self, old_physicians=None, old_physician_rois=None, skip_physicians=False):
        if not skip_physicians:
            self.update_physicians(old_physicians=old_physicians)
        self.update_physician_rois(old_physician_rois=old_physician_rois)
        self.update_variations()
        self.update_uncategorized_ignored_choices()
        self.update_merge_physician_rois()
        self.update_roi_map()

    def on_edit_physician(self, evt):
        current_physicians = list(self.roi_map.get_physicians)
        dlg = RenamePhysicianDialog(self.physician, self.roi_map)
        if dlg.res == wx.ID_OK:
            self.update_all(old_physicians=current_physicians)

    def on_edit_physician_roi(self, evt):
        current_physician_rois = self.roi_map.get_physician_rois(self.physician)
        if self.physician == 'DEFAULT':
            dlg = RenameInstitutionalROIDialog(self.physician_roi, self.roi_map)
        else:
            dlg = RenamePhysicianROIDialog(self.physician, self.physician_roi, self.roi_map)
        if dlg.res == wx.ID_OK:
            self.update_all(old_physician_rois=current_physician_rois, skip_physicians=True)

    def update_physician_enable(self):
        self.button_physician['del'].Enable(self.physician != 'DEFAULT')
        self.button_physician['edit'].Enable(self.physician != 'DEFAULT')
        self.button_variation_add.Enable(self.physician != 'DEFAULT')

    def update_physician_roi_label(self):
        label_text = ['Physician ROI:', 'Institutional ROI:'][self.physician == 'DEFAULT']
        self.label_physician_roi.SetLabelText(label_text)
        self.button_link_physician_roi.Enable(self.physician != 'DEFAULT')

    def uncategorized_ticker(self, evt):
        if self.combo_box_uncategorized_ignored.GetValue() == 'Uncategorized':
            self.button_uncategorized_ignored_ignore.SetLabelText('Ignore DVH')
        else:
            self.button_uncategorized_ignored_ignore.SetLabelText('Unignore DVH')
        self.update_uncategorized_ignored_choices()

    def update_merge_physician_rois(self):
        options = []
        if self.physician != 'DEFAULT':
            options = self.roi_map.get_physician_rois(self.physician)
        if not options:
            options = ['']
        for combo_box in self.combo_box_physician_roi_merge.values():
            combo_box.Clear()
            combo_box.Append(options)
            combo_box.SetValue(options[0])
        self.update_merge_enable()

    @property
    def merge_a(self):
        return self.combo_box_physician_roi_merge['a'].GetValue()

    @property
    def merge_b(self):
        return self.combo_box_physician_roi_merge['b'].GetValue()

    def on_merge(self, evt):
        self.roi_map.merge_physician_rois(self.physician, [self.merge_a, self.merge_b], self.merge_b)
        self.update_all(skip_physicians=True)

    def update_merge_enable(self, *args):  # *args to catch wx.EVT_BUTTON
        self.combo_box_physician_roi_merge['a'].Enable(bool(self.merge_a and self.merge_b))
        self.combo_box_physician_roi_merge['b'].Enable(bool(self.merge_a and self.merge_b))
        self.button_merge.Enable(bool(self.merge_a) and bool(self.merge_b) and self.merge_a != self.merge_b)

    def on_link_physician_roi(self, evt):
        dlg = LinkPhysicianROI(self, self.physician, self.physician_roi, self.roi_map)
        if dlg.res == wx.ID_OK:
            self.update_roi_map()

    def update_tg263_table(self, *evt):
        filter_key_map = {'major': 'Major Cat.', 'minor': 'Minor Cat.',
                          'anatomy': 'Anat. Group', 'target': 'Target Type'}
        data_filter = {col: self.combo_box_tg263[key].GetValue() for key, col in filter_key_map.items()}
        data = self.roi_map_gen.get_filtered_data(data_filter)
        self.data_table_tg263.set_data(data, columns=self.roi_map_gen.keys)
        self.data_table_tg263.set_column_widths(auto=True)

        self.update_tg263_combo_choices()

    def update_tg263_combo_choices(self):
        filter_key_map = {'major': 'Major Cat.', 'minor': 'Minor Cat.',
                          'anatomy': 'Anat. Group', 'target': 'Target Type'}
        for key, col in filter_key_map.items():
            current_choice = self.combo_box_tg263[key].GetValue()
            new_choices = ['All'] + self.data_table_tg263.get_unique_values(col)
            self.combo_box_tg263[key].SetItems(new_choices)
            self.combo_box_tg263[key].SetValue(current_choice)

    def sort_tg263_table(self, evt):
        self.data_table_tg263.sort_table(evt)
        self.data_table_tg263.set_column_widths(auto=True)

    def save_and_update(self, evt):
        RemapROIFrame(self.roi_map)

    def on_cancel(self, *args):
        self.roi_map.import_from_file()
        self.update_roi_map()

    def on_close(self, *args):
        self.Destroy()
        self.roi_map.import_from_file()


class RemapROIWorker(Thread):
    """
    Create a thread to update the SQL DB with roi map changes
    """
    def __init__(self, roi_map, remap_all=False):
        """
        :param roi_map: roi map object
        :type roi_map: DatabaseROIs
        :param remap_all: If true, remap entire database
        :type remap_all: bool
        """
        Thread.__init__(self)

        self.roi_map = roi_map
        self.remap_all = remap_all
        self.start_time = datetime.now()

        self.start()  # start the thread

    def run(self):

        if self.remap_all:
            physician_to_map = list(set(self.roi_map.get_physicians()) - {'DEFAULT'})
            variations_to_update = {}
        else:
            physician_to_map = self.roi_map.physicians_to_remap
            variations_to_update = self.roi_map.variations_to_update

        with DVH_SQL() as cnx:

            physician_counter = 0
            physician_count = len(list(variations_to_update) + physician_to_map)

            # Partial physician remaps
            for physician, variations in variations_to_update.items():
                msg = ["Physician (%s of %s): %s" % (physician_counter+1, physician_count, physician),
                       int(100 * physician_counter / physician_count)]
                wx.CallAfter(pub.sendMessage, "roi_map_update_gauge_1_info", msg=msg)
                physician_counter += 1
                variation_count = len(variations)
                variation_counter = 0

                for variation in variations:
                    msg = ["ROI Name (%s of %s): %s" % (variation_counter+1, variation_count, variation),
                           int(100 * variation_counter / variation_count)]
                    wx.CallAfter(pub.sendMessage, "roi_map_update_gauge_2_info", msg=msg)
                    msg = "Elapsed Time: %s" % get_elapsed_time(self.start_time, datetime.now())
                    wx.CallAfter(pub.sendMessage, "roi_map_update_elapsed_time", msg=msg)
                    variation_counter += 1

                    self.update_variation(variation, physician, cnx)

            # Full Physician remaps
            for physician in physician_to_map:
                condition = "physician = '%s'" % physician
                uids = cnx.get_unique_values('Plans', 'study_instance_uid', condition)
                if uids:
                    condition = "study_instance_uid in ('%s')" % "','".join(uids)
                    variations = cnx.get_unique_values('DVHs', 'roi_name', condition)
                    msg = ["Physician (%s of %s): %s" % (physician_counter+1, physician_count, physician),
                           int(100 * physician_counter / physician_count)]
                    wx.CallAfter(pub.sendMessage, "roi_map_update_gauge_1_info", msg=msg)
                    variation_count = len(variations)
                    variation_counter = 0

                    for variation in variations:
                        variation_frac = variation_counter / variation_count
                        msg = ["Physician (%s of %s): %s" % (physician_counter + 1, physician_count, physician),
                               int(100 * (physician_counter / physician_count + variation_frac / physician_count))]
                        wx.CallAfter(pub.sendMessage, "roi_map_update_gauge_1_info", msg=msg)
                        msg = ["ROI (%s of %s): %s" % (variation_counter+1, variation_count, variation),
                               int(100 * variation_frac)]
                        wx.CallAfter(pub.sendMessage, "roi_map_update_gauge_2_info", msg=msg)
                        msg = "Elapsed Time: %s" % get_elapsed_time(self.start_time, datetime.now())
                        wx.CallAfter(pub.sendMessage, "roi_map_update_elapsed_time", msg=msg)
                        variation_counter += 1

                        self.update_variation(variation, physician, cnx)

                    physician_counter += 1

        wx.CallAfter(pub.sendMessage, "roi_map_close")

    def update_variation(self, variation, physician, cnx):

        new_physician_roi = self.roi_map.get_physician_roi(physician, variation)
        if new_physician_roi == 'uncategorized':
            new_institutional_roi = 'uncategorized'
        else:
            new_institutional_roi = self.roi_map.get_institutional_roi(physician, new_physician_roi)

        condition = "roi_name = '%s'" % variation
        roi_uids = cnx.get_unique_values('DVHs', 'study_instance_uid', condition)
        if roi_uids:
            condition = "physician = '%s' and study_instance_uid in ('%s')" % \
                        (physician, "','".join(roi_uids))
            uids = cnx.get_unique_values('Plans', 'study_instance_uid', condition)

            if uids:
                for i, uid in enumerate(uids):
                    condition = "roi_name = '%s' and study_instance_uid = '%s'" % (variation, uid)
                    cnx.update('dvhs', 'physician_roi', new_physician_roi, condition)
                    cnx.update('dvhs', 'institutional_roi', new_institutional_roi, condition)


class RemapROIFrame(wx.Frame):
    """
    Companion class for RemapROIWorker to to display progress, Automatically calls RemapROIWorker
    """
    def __init__(self, roi_map, remap_all=False):
        """
        :param roi_map: roi map object
        :type roi_map: DatabaseROIs
        :param remap_all: If true, remap entire database
        :type remap_all: bool
        """
        wx.Frame.__init__(self, None, title='Updating Database with ROI Map Changes')
        set_msw_background_color(self)

        self.roi_map = roi_map

        self.gauge_physician = wx.Gauge(self, wx.ID_ANY, 100)
        self.gauge_roi = wx.Gauge(self, wx.ID_ANY, 100)

        self.start_time = None

        self.__set_properties()
        self.__do_layout()
        self.__do_subscribe()

        self.Show()

        RemapROIWorker(self.roi_map, remap_all=remap_all)

    def __set_properties(self):
        self.gauge_physician.SetMinSize((358, 17))
        self.gauge_roi.SetMinSize((358, 17))

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_progress = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL)
        sizer_roi_name = wx.BoxSizer(wx.VERTICAL)
        sizer_physician = wx.BoxSizer(wx.VERTICAL)

        self.label_physician = wx.StaticText(self, wx.ID_ANY, "Physician:")
        sizer_physician.Add(self.label_physician, 0, 0, 0)
        sizer_physician.Add(self.gauge_physician, 0, wx.EXPAND, 0)

        self.label_roi_name = wx.StaticText(self, wx.ID_ANY, "ROI Name:")
        sizer_roi_name.Add(self.label_roi_name, 0, 0, 0)
        sizer_roi_name.Add(self.gauge_roi, 0, wx.EXPAND, 0)

        sizer_progress.Add(sizer_physician, 0, wx.ALL | wx.EXPAND, 5)
        sizer_progress.Add(sizer_roi_name, 0, wx.ALL | wx.EXPAND, 5)
        sizer_wrapper.Add(sizer_progress, 0, wx.ALL | wx.EXPAND, 5)

        self.label_elapsed_time = wx.StaticText(self, wx.ID_ANY, "Elapsed Time:")
        sizer_wrapper.Add(self.label_elapsed_time, 0, wx.BOTTOM | wx.LEFT, 10)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Layout()
        self.Center()

    def __do_subscribe(self):
        pub.subscribe(self.update_gauge_1_info, "roi_map_update_gauge_1_info")
        pub.subscribe(self.update_gauge_2_info, "roi_map_update_gauge_2_info")
        pub.subscribe(self.update_elapsed_time, "roi_map_update_elapsed_time")
        pub.subscribe(self.close, "roi_map_close")

    def close(self):
        self.roi_map.write_to_file()
        self.roi_map.import_from_file()
        self.Destroy()

    def update_gauge_1_info(self, msg):
        wx.CallAfter(self.label_physician.SetLabelText, msg[0])
        wx.CallAfter(self.gauge_physician.SetValue, msg[1])

    def update_gauge_2_info(self, msg):
        wx.CallAfter(self.label_roi_name.SetLabelText, msg[0])
        wx.CallAfter(self.gauge_roi.SetValue, msg[1])

    def update_elapsed_time(self, msg):
        wx.CallAfter(self.label_elapsed_time.SetLabelText, msg)
